@@ -4,16 +4,18 @@ from collections import deque, namedtuple
 from contextlib import contextmanager
 from decimal import Decimal
 from io import BytesIO
+from typing import Any
 from unittest import TestCase, mock
 from urllib.request import pathname2url
 import json
 import os
 import sys
 import tempfile
-import unittest
 import warnings
 
-import attr
+from attrs import define, field
+from referencing.jsonschema import DRAFT202012
+import referencing.exceptions
 
 from jsonschema import (
     FormatChecker,
@@ -22,7 +24,6 @@ from jsonschema import (
     protocols,
     validators,
 )
-from jsonschema.tests._helpers import bug
 
 
 def fail(validator, errors, instance, schema):
@@ -38,6 +39,11 @@ class TestCreateAndExtend(TestCase):
             validators._META_SCHEMAS,
             dict(validators._META_SCHEMAS),
         )
+        self.addCleanup(
+            self.assertEqual,
+            validators._VALIDATORS,
+            dict(validators._VALIDATORS),
+        )
 
         self.meta_schema = {"$id": "some://meta/schema"}
         self.validators = {"fail": fail}
@@ -108,6 +114,7 @@ class TestCreateAndExtend(TestCase):
             version="my version",
         )
         self.addCleanup(validators._META_SCHEMAS.pop, "something")
+        self.addCleanup(validators._VALIDATORS.pop, "my version")
         self.assertEqual(Validator.__name__, "MyVersionValidator")
         self.assertEqual(Validator.__qualname__, "MyVersionValidator")
 
@@ -117,6 +124,7 @@ class TestCreateAndExtend(TestCase):
             version="my version",
         )
         self.addCleanup(validators._META_SCHEMAS.pop, "something")
+        self.addCleanup(validators._VALIDATORS.pop, "my version")
         self.assertEqual(
             repr(Validator({})),
             "MyVersionValidator(schema={}, format_checker=None)",
@@ -128,6 +136,7 @@ class TestCreateAndExtend(TestCase):
             version="my version",
         )
         self.addCleanup(validators._META_SCHEMAS.pop, "something")
+        self.addCleanup(validators._VALIDATORS.pop, "my version")
         self.assertEqual(
             repr(Validator({"a": list(range(1000))})), (
                 "MyVersionValidator(schema={'a': [0, 1, 2, 3, 4, 5, ...]}, "
@@ -148,6 +157,7 @@ class TestCreateAndExtend(TestCase):
             version="foo-bar",
         )
         self.addCleanup(validators._META_SCHEMAS.pop, "something")
+        self.addCleanup(validators._VALIDATORS.pop, "foo-bar")
         self.assertEqual(Validator.__qualname__, "FooBarValidator")
 
     def test_if_a_version_is_not_provided_it_is_not_registered(self):
@@ -165,6 +175,7 @@ class TestCreateAndExtend(TestCase):
             id_of=lambda s: s.get("id", ""),
         )
         self.addCleanup(validators._META_SCHEMAS.pop, meta_schema_key)
+        self.addCleanup(validators._VALIDATORS.pop, "my version")
 
         self.assertIn(meta_schema_key, validators._META_SCHEMAS)
 
@@ -177,6 +188,7 @@ class TestCreateAndExtend(TestCase):
             version="my version",
         )
         self.addCleanup(validators._META_SCHEMAS.pop, meta_schema_key)
+        self.addCleanup(validators._VALIDATORS.pop, "my version")
 
         self.assertIn(meta_schema_key, validators._META_SCHEMAS)
 
@@ -197,6 +209,38 @@ class TestCreateAndExtend(TestCase):
             ),
         )
 
+    def test_check_schema_with_different_metaschema(self):
+        """
+        One can create a validator class whose metaschema uses a different
+        dialect than itself.
+        """
+
+        NoEmptySchemasValidator = validators.create(
+            meta_schema={
+                "$schema": validators.Draft202012Validator.META_SCHEMA["$id"],
+                "not": {"const": {}},
+            },
+        )
+        NoEmptySchemasValidator.check_schema({"foo": "bar"})
+
+        with self.assertRaises(exceptions.SchemaError):
+            NoEmptySchemasValidator.check_schema({})
+
+        NoEmptySchemasValidator({"foo": "bar"}).validate("foo")
+
+    def test_check_schema_with_different_metaschema_defaults_to_self(self):
+        """
+        A validator whose metaschema doesn't declare $schema defaults to its
+        own validation behavior, not the latest "normal" specification.
+        """
+
+        NoEmptySchemasValidator = validators.create(
+            meta_schema={"fail": [{"message": "Meta schema whoops!"}]},
+            validators={"fail": fail},
+        )
+        with self.assertRaises(exceptions.SchemaError):
+            NoEmptySchemasValidator.check_schema({})
+
     def test_extend(self):
         original = dict(self.Validator.VALIDATORS)
         new = object()
@@ -241,6 +285,23 @@ class TestCreateAndExtend(TestCase):
         Derived = validators.extend(Original)
         self.assertEqual(Derived.ID_OF(Derived.META_SCHEMA), correct_id)
 
+    def test_extend_applicable_validators(self):
+        """
+        Extending a validator preserves its notion of applicable validators.
+        """
+
+        schema = {
+            "$defs": {"test": {"type": "number"}},
+            "$ref": "#/$defs/test",
+            "maximum": 1
+        }
+
+        draft4 = validators.Draft4Validator(schema)
+        self.assertTrue(draft4.is_valid(37))  # as $ref ignores siblings
+
+        Derived = validators.extend(validators.Draft4Validator)
+        self.assertTrue(Derived(schema).is_valid(37))
+
 
 class TestValidationErrorMessages(TestCase):
     def message_for(self, instance, schema, *args, **kwargs):
@@ -269,18 +330,6 @@ class TestValidationErrorMessages(TestCase):
         message = self.message_for(instance=1, schema={"type": list(types)})
         self.assertEqual(message, "1 is not of type 'string', 'object'")
 
-    def test_object_without_title_type_failure(self):
-        type = {"type": [{"minimum": 3}]}
-        message = self.message_for(
-            instance=1,
-            schema={"type": [type]},
-            cls=validators.Draft3Validator,
-        )
-        self.assertEqual(
-            message,
-            "1 is not of type {'type': [{'minimum': 3}]}",
-        )
-
     def test_object_with_named_type_failure(self):
         schema = {"type": [{"name": "Foo", "minimum": 3}]}
         message = self.message_for(
@@ -308,6 +357,18 @@ class TestValidationErrorMessages(TestCase):
         )
         self.assertEqual(message, "'foo' is a dependency of 'bar'")
 
+    def test_object_without_title_type_failure_draft3(self):
+        type = {"type": [{"minimum": 3}]}
+        message = self.message_for(
+            instance=1,
+            schema={"type": [type]},
+            cls=validators.Draft3Validator,
+        )
+        self.assertEqual(
+            message,
+            "1 is not of type {'type': [{'minimum': 3}]}",
+        )
+
     def test_dependencies_list_draft3(self):
         depend, on = "bar", "foo"
         schema = {"dependencies": {depend: [on]}}
@@ -449,11 +510,24 @@ class TestValidationErrorMessages(TestCase):
         self.assertEqual(message, "[1, 2, 3] is too long")
 
     def test_prefixItems_with_items(self):
+        message = self.message_for(
+            instance=[1, 2, "foo"],
+            schema={"items": False, "prefixItems": [{}, {}]},
+        )
+        self.assertEqual(
+            message,
+            "Expected at most 2 items but found 1 extra: 'foo'"
+        )
+
+    def test_prefixItems_with_multiple_extra_items(self):
         message = self.message_for(
             instance=[1, 2, "foo", 5],
             schema={"items": False, "prefixItems": [{}, {}]},
         )
-        self.assertEqual(message, "Expected at most 2 items, but found 4")
+        self.assertEqual(
+            message,
+            "Expected at most 2 items but found 2 extra: ['foo', 5]"
+        )
 
     def test_minLength(self):
         message = self.message_for(
@@ -594,7 +668,7 @@ class TestValidationErrorMessages(TestCase):
         message = self.message_for(instance=["foo", "bar"], schema=schema)
         self.assertIn(
             message,
-            "Unevaluated items are not allowed ('bar', 'foo' were unexpected)",
+            "Unevaluated items are not allowed ('foo', 'bar' were unexpected)",
         )
 
     def test_unevaluated_items_on_invalid_type(self):
@@ -602,7 +676,26 @@ class TestValidationErrorMessages(TestCase):
         message = self.message_for(instance="foo", schema=schema)
         self.assertEqual(message, "'foo' is not of type 'array'")
 
-    def test_unevaluated_properties(self):
+    def test_unevaluated_properties_invalid_against_subschema(self):
+        schema = {
+            "properties": {"foo": {"type": "string"}},
+            "unevaluatedProperties": {"const": 12},
+        }
+        message = self.message_for(
+            instance={
+                "foo": "foo",
+                "bar": "bar",
+                "baz": 12,
+            },
+            schema=schema,
+        )
+        self.assertEqual(
+            message,
+            "Unevaluated properties are not valid under the given schema "
+            "('bar' was unevaluated and invalid)",
+        )
+
+    def test_unevaluated_properties_disallowed(self):
         schema = {"type": "object", "unevaluatedProperties": False}
         message = self.message_for(
             instance={
@@ -622,9 +715,82 @@ class TestValidationErrorMessages(TestCase):
         message = self.message_for(instance="foo", schema=schema)
         self.assertEqual(message, "'foo' is not of type 'object'")
 
+    def test_single_item(self):
+        schema = {"prefixItems": [{}], "items": False}
+        message = self.message_for(
+            instance=["foo", "bar", "baz"],
+            schema=schema,
+        )
+        self.assertEqual(
+            message,
+            "Expected at most 1 item but found 2 extra: ['bar', 'baz']",
+        )
+
+    def test_heterogeneous_additionalItems_with_Items(self):
+        schema = {"items": [{}], "additionalItems": False}
+        message = self.message_for(
+            instance=["foo", "bar", 37],
+            schema=schema,
+            cls=validators.Draft7Validator,
+        )
+        self.assertEqual(
+            message,
+            "Additional items are not allowed ('bar', 37 were unexpected)"
+        )
+
+    def test_heterogeneous_items_prefixItems(self):
+        schema = {"prefixItems": [{}], "items": False}
+        message = self.message_for(
+            instance=["foo", "bar", 37],
+            schema=schema,
+        )
+        self.assertEqual(
+            message,
+            "Expected at most 1 item but found 2 extra: ['bar', 37]",
+        )
+
+    def test_heterogeneous_unevaluatedItems_prefixItems(self):
+        schema = {"prefixItems": [{}], "unevaluatedItems": False}
+        message = self.message_for(
+            instance=["foo", "bar", 37],
+            schema=schema,
+        )
+        self.assertEqual(
+            message,
+            "Unevaluated items are not allowed ('bar', 37 were unexpected)",
+        )
+
+    def test_heterogeneous_properties_additionalProperties(self):
+        """
+        Not valid deserialized JSON, but this should not blow up.
+        """
+        schema = {"properties": {"foo": {}}, "additionalProperties": False}
+        message = self.message_for(
+            instance={"foo": {}, "a": "baz", 37: 12},
+            schema=schema,
+        )
+        self.assertEqual(
+            message,
+            "Additional properties are not allowed (37, 'a' were unexpected)",
+        )
+
+    def test_heterogeneous_properties_unevaluatedProperties(self):
+        """
+        Not valid deserialized JSON, but this should not blow up.
+        """
+        schema = {"properties": {"foo": {}}, "unevaluatedProperties": False}
+        message = self.message_for(
+            instance={"foo": {}, "a": "baz", 37: 12},
+            schema=schema,
+        )
+        self.assertEqual(
+            message,
+            "Unevaluated properties are not allowed (37, 'a' were unexpected)",
+        )
+
 
 class TestValidationErrorDetails(TestCase):
-    # TODO: These really need unit tests for each individual validator, rather
+    # TODO: These really need unit tests for each individual keyword, rather
     #       than just these higher level tests.
     def test_anyOf(self):
         instance = 5
@@ -1127,7 +1293,7 @@ class TestValidationErrorDetails(TestCase):
         ref, schema = "someRef", {"additionalProperties": {"type": "integer"}}
         validator = validators.Draft7Validator(
             {"$ref": ref},
-            resolver=validators.RefResolver("", {}, store={ref: schema}),
+            resolver=validators._RefResolver("", {}, store={ref: schema}),
         )
         error, = validator.iter_errors({"foo": "notAnInteger"})
 
@@ -1406,11 +1572,11 @@ class TestValidationErrorDetails(TestCase):
         )
 
 
-class MetaSchemaTestsMixin(object):
+class MetaSchemaTestsMixin:
     # TODO: These all belong upstream
     def test_invalid_properties(self):
         with self.assertRaises(exceptions.SchemaError):
-            self.Validator.check_schema({"properties": {"test": object()}})
+            self.Validator.check_schema({"properties": 12})
 
     def test_minItems_invalid_string(self):
         with self.assertRaises(exceptions.SchemaError):
@@ -1421,20 +1587,48 @@ class MetaSchemaTestsMixin(object):
         """
         Technically, all the spec says is they SHOULD have elements, not MUST.
 
+        (As of Draft 6. Previous drafts do say MUST).
+
         See #529.
         """
-        self.Validator.check_schema({"enum": []})
+        if self.Validator in {
+            validators.Draft3Validator,
+            validators.Draft4Validator,
+        }:
+            with self.assertRaises(exceptions.SchemaError):
+                self.Validator.check_schema({"enum": []})
+        else:
+            self.Validator.check_schema({"enum": []})
 
     def test_enum_allows_non_unique_items(self):
         """
         Technically, all the spec says is they SHOULD be unique, not MUST.
 
+        (As of Draft 6. Previous drafts do say MUST).
+
         See #529.
         """
-        self.Validator.check_schema({"enum": [12, 12]})
+        if self.Validator in {
+            validators.Draft3Validator,
+            validators.Draft4Validator,
+        }:
+            with self.assertRaises(exceptions.SchemaError):
+                self.Validator.check_schema({"enum": [12, 12]})
+        else:
+            self.Validator.check_schema({"enum": [12, 12]})
+
+    def test_schema_with_invalid_regex(self):
+        with self.assertRaises(exceptions.SchemaError):
+            self.Validator.check_schema({"pattern": "*notaregex"})
+
+    def test_schema_with_invalid_regex_with_disabled_format_validation(self):
+        self.Validator.check_schema(
+            {"pattern": "*notaregex"},
+            format_checker=None,
+        )
 
 
-class ValidatorTestMixin(MetaSchemaTestsMixin, object):
+class ValidatorTestMixin(MetaSchemaTestsMixin):
     def test_it_implements_the_validator_protocol(self):
         self.assertIsInstance(self.Validator({}), protocols.Validator)
 
@@ -1449,37 +1643,48 @@ class ValidatorTestMixin(MetaSchemaTestsMixin, object):
     def test_non_existent_properties_are_ignored(self):
         self.Validator({object(): object()}).validate(instance=object())
 
-    def test_it_creates_a_ref_resolver_if_not_provided(self):
-        self.assertIsInstance(
-            self.Validator({}).resolver,
-            validators.RefResolver,
+    def test_evolve(self):
+        schema, format_checker = {"type": "integer"}, FormatChecker()
+        original = self.Validator(
+            schema,
+            format_checker=format_checker,
+        )
+        new = original.evolve(
+            schema={"type": "string"},
+            format_checker=self.Validator.FORMAT_CHECKER,
         )
 
-    def test_it_delegates_to_a_ref_resolver(self):
-        ref, schema = "someCoolRef", {"type": "integer"}
-        resolver = validators.RefResolver("", {}, store={ref: schema})
-        validator = self.Validator({"$ref": ref}, resolver=resolver)
+        expected = self.Validator(
+            {"type": "string"},
+            format_checker=self.Validator.FORMAT_CHECKER,
+            _resolver=new._resolver,
+        )
 
-        with self.assertRaises(exceptions.ValidationError):
-            validator.validate(None)
+        self.assertEqual(new, expected)
+        self.assertNotEqual(new, original)
 
-    def test_it_delegates_to_a_legacy_ref_resolver(self):
+    def test_evolve_with_subclass(self):
         """
-        Legacy RefResolvers support only the context manager form of
-        resolution.
+        Subclassing validators isn't supported public API, but some users have
+        done it, because we don't actually error entirely when it's done :/
+
+        We need to deprecate doing so first to help as many of these users
+        ensure they can move to supported APIs, but this test ensures that in
+        the interim, we haven't broken those users.
         """
 
-        class LegacyRefResolver(object):
-            @contextmanager
-            def resolving(this, ref):
-                self.assertEqual(ref, "the ref")
-                yield {"type": "integer"}
+        with self.assertWarns(DeprecationWarning):
+            @define
+            class OhNo(self.Validator):
+                foo = field(factory=lambda: [1, 2, 3])
+                _bar = field(default=37)
 
-        resolver = LegacyRefResolver()
-        schema = {"$ref": "the ref"}
+        validator = OhNo({}, bar=12)
+        self.assertEqual(validator.foo, [1, 2, 3])
 
-        with self.assertRaises(exceptions.ValidationError):
-            self.Validator(schema, resolver=resolver).validate(None)
+        new = validator.evolve(schema={"type": "integer"})
+        self.assertEqual(new.foo, [1, 2, 3])
+        self.assertEqual(new._bar, 12)
 
     def test_is_type_is_true_for_valid_type(self):
         self.assertTrue(self.Validator({}).is_type("foo", "string"))
@@ -1533,7 +1738,7 @@ class ValidatorTestMixin(MetaSchemaTestsMixin, object):
             elif value == "bad":
                 raise bad
             else:  # pragma: no cover
-                self.fail("What is {}? [Baby Don't Hurt Me]".format(value))
+                self.fail(f"What is {value}? [Baby Don't Hurt Me]")
 
         validator = self.Validator(
             {"format": "foo"}, format_checker=checker,
@@ -1641,8 +1846,39 @@ class ValidatorTestMixin(MetaSchemaTestsMixin, object):
             with self.assertRaises(exceptions.ValidationError):
                 validator.validate(instance)
 
+    def test_it_creates_a_ref_resolver_if_not_provided(self):
+        with self.assertWarns(DeprecationWarning):
+            resolver = self.Validator({}).resolver
+        self.assertIsInstance(resolver, validators._RefResolver)
 
-class AntiDraft6LeakMixin(object):
+    def test_it_upconverts_from_deprecated_RefResolvers(self):
+        ref, schema = "someCoolRef", {"type": "integer"}
+        resolver = validators._RefResolver("", {}, store={ref: schema})
+        validator = self.Validator({"$ref": ref}, resolver=resolver)
+
+        with self.assertRaises(exceptions.ValidationError):
+            validator.validate(None)
+
+    def test_it_upconverts_from_yet_older_deprecated_legacy_RefResolvers(self):
+        """
+        Legacy RefResolvers support only the context manager form of
+        resolution.
+        """
+
+        class LegacyRefResolver:
+            @contextmanager
+            def resolving(this, ref):
+                self.assertEqual(ref, "the ref")
+                yield {"type": "integer"}
+
+        resolver = LegacyRefResolver()
+        schema = {"$ref": "the ref"}
+
+        with self.assertRaises(exceptions.ValidationError):
+            self.Validator(schema, resolver=resolver).validate(None)
+
+
+class AntiDraft6LeakMixin:
     """
     Make sure functionality from draft 6 doesn't leak backwards in time.
     """
@@ -1657,18 +1893,14 @@ class AntiDraft6LeakMixin(object):
             self.Validator.check_schema(False)
         self.assertIn("False is not of type", str(e.exception))
 
-    @unittest.skip(bug(523))
     def test_True_is_not_a_schema_even_if_you_forget_to_check(self):
-        resolver = validators.RefResolver("", {})
         with self.assertRaises(Exception) as e:
-            self.Validator(True, resolver=resolver).validate(12)
+            self.Validator(True).validate(12)
         self.assertNotIsInstance(e.exception, exceptions.ValidationError)
 
-    @unittest.skip(bug(523))
     def test_False_is_not_a_schema_even_if_you_forget_to_check(self):
-        resolver = validators.RefResolver("", {})
         with self.assertRaises(Exception) as e:
-            self.Validator(False, resolver=resolver).validate(12)
+            self.Validator(False).validate(12)
         self.assertNotIsInstance(e.exception, exceptions.ValidationError)
 
 
@@ -1734,6 +1966,21 @@ class TestDraft202012Validator(ValidatorTestMixin, TestCase):
     invalid = {"type": "integer"}, "foo"
 
 
+class TestLatestValidator(TestCase):
+    """
+    These really apply to multiple versions but are easiest to test on one.
+    """
+
+    def test_ref_resolvers_may_have_boolean_schemas_stored(self):
+        ref = "someCoolRef"
+        schema = {"$ref": ref}
+        resolver = validators._RefResolver("", {}, store={ref: False})
+        validator = validators._LATEST_VERSION(schema, resolver=resolver)
+
+        with self.assertRaises(exceptions.ValidationError):
+            validator.validate(None)
+
+
 class TestValidatorFor(TestCase):
     def test_draft_3(self):
         schema = {"$schema": "http://json-schema.org/draft-03/schema"}
@@ -1872,6 +2119,10 @@ class TestValidatorFor(TestCase):
             validators.validator_for(schema={}, default={})
         self.assertFalse(w)
 
+    def test_validator_for_custom_default_with_schema(self):
+        schema, default = {"$schema": "mailto:foo@example.com"}, object()
+        self.assertIs(validators.validator_for(schema, default), default)
+
 
 class TestValidate(TestCase):
     def assertUses(self, schema, Validator):
@@ -1959,11 +2210,72 @@ class TestValidate(TestCase):
         )
 
     def test_it_uses_best_match(self):
-        # This is a schema that best_match will recurse into
-        schema = {"oneOf": [{"type": "string"}, {"type": "array"}]}
+        schema = {
+            "oneOf": [
+                {"type": "number", "minimum": 20},
+                {"type": "array"},
+            ],
+        }
         with self.assertRaises(exceptions.ValidationError) as e:
             validators.validate(12, schema)
-        self.assertIn("12 is not of type", str(e.exception))
+        self.assertIn("12 is less than the minimum of 20", str(e.exception))
+
+
+class TestThreading(TestCase):
+    """
+    Threading-related functionality tests.
+
+    jsonschema doesn't promise thread safety, and its validation behavior
+    across multiple threads may change at any time, but that means it isn't
+    safe to share *validators* across threads, not that anytime one has
+    multiple threads that jsonschema won't work (it certainly is intended to).
+
+    These tests ensure that this minimal level of functionality continues to
+    work.
+    """
+
+    def test_validation_across_a_second_thread(self):
+        failed = []
+
+        def validate():
+            try:
+                validators.validate(instance=37, schema=True)
+            except:  # pragma: no cover  # noqa: E722
+                failed.append(sys.exc_info())
+
+        validate()  # just verify it succeeds
+
+        from threading import Thread
+        thread = Thread(target=validate)
+        thread.start()
+        thread.join()
+        self.assertEqual((thread.is_alive(), failed), (False, []))
+
+
+class TestReferencing(TestCase):
+    def test_registry_with_retrieve(self):
+        def retrieve(uri):
+            return DRAFT202012.create_resource({"type": "integer"})
+
+        registry = referencing.Registry(retrieve=retrieve)
+        schema = {"$ref": "https://example.com/"}
+        validator = validators.Draft202012Validator(schema, registry=registry)
+
+        self.assertEqual(
+            (validator.is_valid(12), validator.is_valid("foo")),
+            (True, False),
+        )
+
+    def test_custom_registries_do_not_autoretrieve_remote_resources(self):
+        registry = referencing.Registry()
+        schema = {"$ref": "https://example.com/"}
+        validator = validators.Draft202012Validator(schema, registry=registry)
+
+        with warnings.catch_warnings(record=True) as w:
+            warnings.simplefilter("always")
+            with self.assertRaises(referencing.exceptions.Unresolvable):
+                validator.validate(12)
+        self.assertFalse(w)
 
 
 class TestRefResolver(TestCase):
@@ -1975,7 +2287,7 @@ class TestRefResolver(TestCase):
     def setUp(self):
         self.referrer = {}
         self.store = {self.stored_uri: self.stored_schema}
-        self.resolver = validators.RefResolver(
+        self.resolver = validators._RefResolver(
             self.base_uri, self.referrer, self.store,
         )
 
@@ -1995,7 +2307,7 @@ class TestRefResolver(TestCase):
 
     def test_it_resolves_local_refs_with_id(self):
         schema = {"id": "http://bar/schema#", "a": {"foo": "bar"}}
-        resolver = validators.RefResolver.from_schema(
+        resolver = validators._RefResolver.from_schema(
             schema,
             id_of=lambda schema: schema.get("id", ""),
         )
@@ -2016,7 +2328,7 @@ class TestRefResolver(TestCase):
         ref = "http://bar#baz"
         schema = {"baz": 12}
 
-        if "requests" in sys.modules:
+        if "requests" in sys.modules:  # pragma: no cover
             self.addCleanup(
                 sys.modules.__setitem__, "requests", sys.modules["requests"],
             )
@@ -2029,7 +2341,7 @@ class TestRefResolver(TestCase):
         ref = "http://bar#baz"
         schema = {"baz": 12}
 
-        if "requests" in sys.modules:
+        if "requests" in sys.modules:  # pragma: no cover
             self.addCleanup(
                 sys.modules.__setitem__, "requests", sys.modules["requests"],
             )
@@ -2052,13 +2364,13 @@ class TestRefResolver(TestCase):
             self.addCleanup(os.remove, tempf.name)
             json.dump({"foo": "bar"}, tempf)
 
-        ref = "file://{}#foo".format(pathname2url(tempf.name))
+        ref = f"file://{pathname2url(tempf.name)}#foo"
         with self.resolver.resolving(ref) as resolved:
             self.assertEqual(resolved, "bar")
 
     def test_it_can_construct_a_base_uri_from_a_schema(self):
         schema = {"id": "foo"}
-        resolver = validators.RefResolver.from_schema(
+        resolver = validators._RefResolver.from_schema(
             schema,
             id_of=lambda schema: schema.get("id", ""),
         )
@@ -2075,7 +2387,7 @@ class TestRefResolver(TestCase):
 
     def test_it_can_construct_a_base_uri_from_a_schema_without_id(self):
         schema = {}
-        resolver = validators.RefResolver.from_schema(schema)
+        resolver = validators._RefResolver.from_schema(schema)
         self.assertEqual(resolver.base_uri, "")
         self.assertEqual(resolver.resolution_scope, "")
         with resolver.resolving("") as resolved:
@@ -2090,7 +2402,7 @@ class TestRefResolver(TestCase):
 
         schema = {"foo": "bar"}
         ref = "foo://bar"
-        resolver = validators.RefResolver("", {}, handlers={"foo": handler})
+        resolver = validators._RefResolver("", {}, handlers={"foo": handler})
         with resolver.resolving(ref) as resolved:
             self.assertEqual(resolved, schema)
 
@@ -2104,7 +2416,7 @@ class TestRefResolver(TestCase):
                 self.fail("Response must not have been cached!")
 
         ref = "foo://bar"
-        resolver = validators.RefResolver(
+        resolver = validators._RefResolver(
             "", {}, cache_remote=True, handlers={"foo": handler},
         )
         with resolver.resolving(ref):
@@ -2122,7 +2434,7 @@ class TestRefResolver(TestCase):
                 self.fail("Handler called twice!")
 
         ref = "foo://bar"
-        resolver = validators.RefResolver(
+        resolver = validators._RefResolver(
             "", {}, cache_remote=False, handlers={"foo": handler},
         )
         with resolver.resolving(ref):
@@ -2135,19 +2447,76 @@ class TestRefResolver(TestCase):
             raise error
 
         ref = "foo://bar"
-        resolver = validators.RefResolver("", {}, handlers={"foo": handler})
-        with self.assertRaises(exceptions.RefResolutionError) as err:
+        resolver = validators._RefResolver("", {}, handlers={"foo": handler})
+        with self.assertRaises(exceptions._RefResolutionError) as err:
             with resolver.resolving(ref):
                 self.fail("Shouldn't get this far!")  # pragma: no cover
-        self.assertEqual(err.exception, exceptions.RefResolutionError(error))
+        self.assertEqual(err.exception, exceptions._RefResolutionError(error))
 
     def test_helpful_error_message_on_failed_pop_scope(self):
-        resolver = validators.RefResolver("", {})
+        resolver = validators._RefResolver("", {})
         resolver.pop_scope()
-        with self.assertRaises(exceptions.RefResolutionError) as exc:
+        with self.assertRaises(exceptions._RefResolutionError) as exc:
             resolver.pop_scope()
         self.assertIn("Failed to pop the scope", str(exc.exception))
 
+    def test_pointer_within_schema_with_different_id(self):
+        """
+        See #1085.
+        """
+        schema = validators.Draft7Validator.META_SCHEMA
+        one = validators._RefResolver("", schema)
+        validator = validators.Draft7Validator(schema, resolver=one)
+        self.assertFalse(validator.is_valid({"maxLength": "foo"}))
+
+        another = {
+            "allOf": [{"$ref": validators.Draft7Validator.META_SCHEMA["$id"]}],
+        }
+        two = validators._RefResolver("", another)
+        validator = validators.Draft7Validator(another, resolver=two)
+        self.assertFalse(validator.is_valid({"maxLength": "foo"}))
+
+    def test_newly_created_validator_with_ref_resolver(self):
+        """
+        See https://github.com/python-jsonschema/jsonschema/issues/1061#issuecomment-1624266555.
+        """
+
+        def handle(uri):
+            self.assertEqual(uri, "http://example.com/foo")
+            return {"type": "integer"}
+
+        resolver = validators._RefResolver("", {}, handlers={"http": handle})
+        Validator = validators.create(
+            meta_schema={},
+            validators=validators.Draft4Validator.VALIDATORS,
+        )
+        schema = {"$id": "http://example.com/bar", "$ref": "foo"}
+        validator = Validator(schema, resolver=resolver)
+        self.assertEqual(
+            (validator.is_valid({}), validator.is_valid(37)),
+            (False, True),
+        )
+
+    def test_refresolver_with_pointer_in_schema_with_no_id(self):
+        """
+        See https://github.com/python-jsonschema/jsonschema/issues/1124#issuecomment-1632574249.
+        """
+
+        schema = {
+            "properties": {"x": {"$ref": "#/definitions/x"}},
+            "definitions": {"x": {"type": "integer"}},
+        }
+
+        validator = validators.Draft202012Validator(
+            schema,
+            resolver=validators._RefResolver("", schema),
+        )
+        self.assertEqual(
+            (validator.is_valid({"x": "y"}), validator.is_valid({"x": 37})),
+            (False, True),
+        )
+
+
 
 def sorted_errors(errors):
     def key(error):
@@ -2158,10 +2527,10 @@ def sorted_errors(errors):
     return sorted(errors, key=key)
 
 
-@attr.s
-class ReallyFakeRequests(object):
+@define
+class ReallyFakeRequests:
 
-    _responses = attr.ib()
+    _responses: dict[str, Any]
 
     def get(self, url):
         response = self._responses.get(url)
@@ -2170,10 +2539,10 @@ class ReallyFakeRequests(object):
         return _ReallyFakeJSONResponse(json.dumps(response))
 
 
-@attr.s
-class _ReallyFakeJSONResponse(object):
+@define
+class _ReallyFakeJSONResponse:
 
-    _response = attr.ib()
+    _response: str
 
     def json(self):
         return json.loads(self._response)
